/******************************************************************************
* License of Lua:
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
******************************************************************************/

#include <os.h>
#include <lauxlib.h>


// Special definitions for the Lua functions - we can't use syscalls in our hook.

// LUA_API int lua_gettop (lua_State *L);
#undef lua_gettop
static const int lua_gettop_addrs[] = {0X106C4FCC, 0X106A94C0, 0X106A5FB8, 0X106A6754};
#define lua_gettop SYSCALL_CUSTOM(lua_gettop_addrs, int, lua_State *)

// LUA_API void lua_getfield (lua_State *L, int idx, const char *k);
#undef lua_getfield
static const int lua_getfield_addrs[] = {0X106C60BC, 0X106AA5B0, 0X106A70A8, 0X106A7844};
#define lua_getfield SYSCALL_CUSTOM(lua_getfield_addrs, int, lua_State *, int, const char *)

// LUA_API void lua_pushvalue (lua_State *L, int idx);
#undef lua_pushvalue
static const int lua_pushvalue_addrs[] = {0X106C513C, 0X106A9630, 0X106A6128, 0X106A68C4};
#define lua_pushvalue SYSCALL_CUSTOM(lua_pushvalue_addrs, int, lua_State *, int)

// LUA_API void lua_call (lua_State *L, int nargs, int nresults);
#undef lua_call
static const int lua_call_addrs[] = {0X106C5D00, 0X106AA1F4, 0X106A6CEC, 0X106A7488};
#define lua_call SYSCALL_CUSTOM(lua_call_addrs, int, lua_State *, int, int)

// LUA_API const char *lua_tolstring (lua_State *L, int idx, size_t *len);
#undef lua_tolstring
static const int lua_tolstring_addrs[] = {0x106C624C, 0X106AA740, 0X106A7238, 0X106A79D4};
#define lua_tolstring SYSCALL_CUSTOM(lua_tolstring_addrs, const char *, lua_State *, int, size_t *)

// LUALIB_API int luaL_error (lua_State *L, const char *fmt, ...);
#undef luaL_error
static const int luaL_error_addrs[] = {0X106C7294, 0X106AB788, 0X106A8280, 0X106A8A1C};
// Lowering a varargs syscall to a fixedargs syscall.
#define luaL_error SYSCALL_CUSTOM(luaL_error_addrs, int, lua_State *, const char *)

// LUA_API void lua_settop (lua_State *L, int idx);
#undef lua_settop
static const int lua_settop_addrs[] = {0X106C4FE0, 0X106A94D4, 0X106A5FCC, 0X106A6768};
#define lua_settop SYSCALL_CUSTOM(lua_settop_addrs, int, lua_State *, int)


//! The array defining the base Lua functions
static const int base_funcs_addrs[] = {0X108DBF00, 0X10916DF8, 0X108B7B30, 0X10913EB0};
#define base_funcs ((struct luaL_Reg *)nl_osvalue((int*)base_funcs_addrs, sizeof(base_funcs_addrs)/sizeof(base_funcs_addrs[0])))


//! The address of the original luaB_print function
static const int orig_luaB_print_addrs[] = {0X106C7948, 0X106ABE3C, 0X106A8934, 0x106A90D0};
#define orig_luaB_print ((lua_CFunction)nl_osvalue((int*)orig_luaB_print_addrs, sizeof(base_funcs_addrs)/sizeof(base_funcs_addrs[0])))


//! The function behind print() in standard Lua, extracted from Lua 5.1.4 (the version used by TI)
//! and slightly modified: fputs(X, stdout) -> printf(X).
static int luaB_print (lua_State *L) {
    int n;
    int i;
    n = lua_gettop(L);  /* number of arguments */

    lua_getglobal(L, "tostring");
    for (i=1; i<=n; i++) {
        const char *s;
        lua_pushvalue(L, -1);  /* function to be called */
        lua_pushvalue(L, i);   /* value to print */
        lua_call(L, 1, 1);
        s = lua_tostring(L, -1);  /* get result */
        if (s == NULL) {
            return luaL_error(L, LUA_QL("tostring") " must return a string to " LUA_QL("print"));
        }
        if (i>1) {
            printf("\t");
        }
        printf(s);
        lua_pop(L, 1);  /* pop result */
    }
    printf("\n");
    return 0;
}


//! Guess what, a main is a main !
int main(void) {
    if (base_funcs[10].func != orig_luaB_print) {
        printf("The print function is already overwritten, bailing out\n");
        return 0;
    }

    printf("Overwriting the buggy print function\n");
    base_funcs[10].func = luaB_print;
    // Tell Ndless we are a TSR, so that it doesn't clean us up.
    nl_set_resident();
    // We're done - as simple as that :)
    return 0;
}
